// ***********************************************************
// Project: lcd HD44780 4-bit
// Author: StartCd.narod.ru
// : 1.4 02.04.2009
// c   bv 
//   LCD   
// ***********************************************************
#include <avr/io.h>			// Most basic include files
#include <avr/interrupt.h>	// Add the necessary ones
#include <util/delay.h>		/*F_CPU = 4000000; _delay_ms max = 65,535ms; _delay_us max = 192us */
#include <avr/pgmspace.h>

#include "lph.h"


static void lcd_4bus_io(uint8_t data){
	OFF(DB7);
	OFF(DB6);
	OFF(DB5);
	OFF(DB4);

	ON(E);
	if (data&_BV(7)) ON(DB7);
	if (data&_BV(6)) ON(DB6);
	if (data&_BV(5)) ON(DB5);
	if (data&_BV(4)) ON(DB4);
	_delay_us(3);
	OFF(E);
}

void lcd_4bus_write(uint8_t data)	//   4  
{	//  
	DRIVER(DB7,OUT);
	DRIVER(DB6,OUT);
	DRIVER(DB5,OUT);
	DRIVER(DB4,OUT);
	
	lcd_4bus_io(data);
	lcd_4bus_io(data<<4);
	
	OFF(DB7);
	OFF(DB6);
	OFF(DB5);
	OFF(DB4);
	DRIVER(DB7,IN);
	DRIVER(DB6,IN);
	DRIVER(DB5,IN);
	DRIVER(DB4,IN);

}

uint8_t lcd_read_byte(void){//   4   
uint8_t byte=0;
	OFF(DB7);
	OFF(DB6);
	OFF(DB5);
	OFF(DB4);
	DRIVER(DB7,IN);
	DRIVER(DB6,IN);
	DRIVER(DB5,IN);
	DRIVER(DB4,IN);
	
	OFF(RS);
	ON(RW);

	ON(E);
	_delay_us(3);
	if ( ACTIVE(DB7) ) byte|=_BV(7);
	if ( ACTIVE(DB6) ) byte|=_BV(6);
	if ( ACTIVE(DB5) ) byte|=_BV(5);
	if ( ACTIVE(DB4) ) byte|=_BV(4);
	OFF(E);
	
	_delay_us(1);

	ON(E);
	_delay_us(3);
	if ( ACTIVE(DB7) ) byte|=_BV(3);
	if ( ACTIVE(DB6) ) byte|=_BV(2);
	if ( ACTIVE(DB5) ) byte|=_BV(1);
	if ( ACTIVE(DB4) ) byte|=_BV(0);
	OFF(E);

	OFF(RW);
	return byte;
	//  , RS=RW=0,  DB3..0 = 0    Z  
}

uint8_t read_busy_address(void){
uint8_t byte;

	while( (uint8_t) (( byte=lcd_read_byte() ) & busy_flag) );//   busy_flag
	return byte;
	//   AC, RS=RW=0,  DB3..0 = 0    Z  
}

void lcd_write_data(uint8_t data){
	//write data to CGRAM or DDRAM
	read_busy_address();
	ON(RS);//RS = 1 -   (DR). RW = 0 - 
	lcd_4bus_write(data);
	OFF(RS);
}



void lcd_instructions_set(uint8_t address){
	//   -> address|( , cgram  ddram)
	read_busy_address();
	lcd_4bus_write(address);
}

void lcd_set_address(uint8_t x, uint8_t y){
	x|=set_ddram_address;
	if (y) x|=0x40;
	lcd_instructions_set(x);
}

void lcd_init(void){
	
	DRIVER(E,OUT);	
	DRIVER(RW,OUT);	
	DRIVER(RS,OUT);	

	DRIVER(DB7,OUT);
	DRIVER(DB6,OUT);
	DRIVER(DB5,OUT);
	DRIVER(DB4,OUT);

	OFF(E);
	OFF(RW);
	OFF(RS);
	OFF(DB7);
	OFF(DB6);
	OFF(DB5);
	OFF(DB4);
	
	_delay_ms(40);	//   20 (   5)
	
	ON(E);			//-----------1--------------
	ON(DB5);		//  $30 
	ON(DB4);	
	OFF(E);			//-------------------------
		
	_delay_ms(5);	//  2ms
	
	ON(E);			//-----------2--------------
	//ON(DB5);		//  $30 
	//ON(DB4);
	_delay_us(3);
	OFF(E);			//-------------------------
	
	_delay_us(110);//  100us
	
	ON(E);			//-----------3--------------
	//ON(DB5);		//  $30.          
	//ON(DB4);		//     (       8-  ) 
	_delay_us(3);	//   .
	OFF(E);			//-------------------------
	
	_delay_us(50);	//40us
	
	OFF(DB4);		//D/L:  ,    : 0 - 4 , 1 - 8 .
	ON(E);			 
	//ON(DB5);
	_delay_us(3);			
	OFF(E);
	
	lcd_instructions_set(0x28);//0x28 - 2 , 5  8 . (0x2C - 2  510)
	lcd_instructions_set(0x08);// : D e - ,
	lcd_instructions_set(0x06);//    =1 - .
	lcd_instructions_set(clear_display);

	lcd_instructions_set(cursor_off);//   

}

//****************************************************************************************************

void LCDwritebyte(uint8_t a, uint8_t pc)//      LCD
	{
	lcd_instructions_set(a);
	lcd_write_data(pc);
	}
	
void LCDdefinechar(const uint8_t *pc,uint8_t char_code){ //    FLASHROM
	uint8_t a, pcc;
	uint16_t i;
	a=(char_code<<3)|0x40;
	for (i=0; i<8; i++){
		pcc=pgm_read_byte(&pc[i]);
		LCDwritebyte(a++,pcc);
		}
	}
	
void lcd_puts_p(const char *progmem_str )
{
    uint8_t c;
    while ( (c = pgm_read_byte(progmem_str++)) ) 
      lcd_write_data(c);
}


